<?php
/* ********************************************************************
* Interleave
* Copyright (c) 2001-2016 Atomos Applications The Netherlands - www.atomos.nl - info@atomos.nl - www.interleave.nl - info@interleave.nl
* Licensed under the GNU GPL. For full terms see http://www.gnu.org/
*
* This file contains all functions related to the handling of rules
*
* Check http://www.interleave.nl/ for more information
*
* BUILD / RELEASE :: 7.0.20161227
*
**********************************************************************
*/

// aantal nieuwe conditions en actions moet nog support voor ingebouwd worden
/*
FAILED CONDITIONS
I: Start evaluation of rule '105' named 'Converted from triggers'.
I: Table 'entity', record '16014', field '', condition 'changed'
I: Operator for conditions is 'AND'
F: Condition '[value] days before [field]' is invalid.

FOUT!! Fixen


*/

function Processrules($triggering_table, $triggering_record, $triggering_field, $triggering_condition, $new_value, $current_record=false, $log_fails = true) {

	if (db_GetValue("SELECT COUNT(*) FROM `CRMrules`") == 0) {
		return true;
	}

	/* when in development, log fails and passes regardless of the rule configuration */
	$log_because_of_development = false;
	if (GetSetting("DevelopmentOrProduction") == "Development") {
		$log_because_of_development = true;
	}

	//print ("<h2>Processrules($triggering_table, $triggering_record, $triggering_field, $triggering_condition, $new_value) </h2>");

	// The System data structure MUST be used because rules bypass security. The System data structure is not loaded by default.
	if (!is_array($GLOBALS['SystemDataStructure'])) {
		$GLOBALS['SystemDataStructure'] = GetSystemDataStructure();
	}

	// SLORDIG en memory intensief; Aan Jeroen vragen of dit netter kan.

	// Bypass PDS security by overwriting it to SDS
	/* dont park the user's PDS in the gloval scope! this function could kick a module, which invokes this function again and
	   as this function ends with an unset of the parkted PDS, we'll lose the PDS alltogether */

	$PDSPARK = $GLOBALS['UC']['PersonalDataStructure'];
	$GLOBALS['UC']['PersonalDataStructure'] = $GLOBALS['SystemDataStructure'];

	$record = array();

	// If a record is given, fetch it, but use system rights
	if (ipi($triggering_record)) {

		if (!$current_record) {

			$fields_to_fetch = array();

			$sql = "SELECT `CRMruleconditions`.`field` FROM `CRMruleconditions` WHERE `CRMruleconditions`.`field` LIKE '" . $triggering_table . "-%' UNION SELECT `CRMruleactionconditions`.`field` FROM `CRMruleactionconditions` WHERE `CRMruleactionconditions`.`field` LIKE '" . $triggering_table . "-%' UNION SELECT `CRMruleactions`.`field` FROM `CRMruleactions` WHERE `CRMruleactions`.`field` LIKE '" . $triggering_table . "-%'";

			/* now add all primaries to the fields to fetch */
			$fields_to_fetch_tmp = db_GetFlatArray($sql);
			foreach ($fields_to_fetch_tmp AS $ftf) {

				/* add current always */
				$fields_to_fetch[] = $ftf;

				$elements = explode("-", $ftf);
				$element_table = $elements[count($elements)-2];

				/* last table is not our own table, add primary to query */
				if ($element_table != $triggering_table) {

					unset($elements[count($elements)-1]);

					$path = implode("-", $elements) . "-" . $GLOBALS['SystemDataStructure'][$element_table]["primary"];

					if (!in_array($path, $fields_to_fetch)) {
						$fields_to_fetch[] = $path;
					}
				}
			}

			$selectallfields = false;
			$bypass_record_access_by_sql = false;

			// Get the record. It will be passed along.

			$record = getrecord($triggering_table, $triggering_record, $selectallfields, $bypass_record_access_by_sql, $fields_to_fetch);

			//trace("fetched record myself");
		} else {
			//trace("got record from caller");
			$record = &$current_record;
		}
	}
	$add_sql = "";

	if ($triggering_field != "" && $triggering_condition == "changed") {

		/* accelerate this to make it faster - a lot - by excluding all rules we're sure are not matching the criteria. we could do more at this point */

		$add_sql .= " AND `CRMrules`.`id` IN (SELECT `CRMruleconditions`.`rule` FROM `CRMruleconditions` WHERE `CRMruleconditions`.`condition` = '[field] must be changed' AND `CRMruleconditions`.`field`='" . $triggering_table . "-" . mres($triggering_field) . "')";

	}

	// Get the (enabled) rules for this table
	$sql = "SELECT * FROM `CRMrules` WHERE `CRMrules`.`table`='" . mres($triggering_table) . "' AND `CRMrules`.`enabled`='Yes' " . $add_sql . " ORDER BY `CRMrules`.`id`";
//	trace($sql);
	$rules = db_GetArray($sql);

	//if ($GLOBALS['NIGHTLY_CRON']) print date('r') . ": Table " . $triggering_table . ", num of rules " . count($rules) . "\n";

	// Cycle through all rules
	foreach ($rules AS $rule) {

	//	if ($GLOBALS['NIGHTLY_CRON']) print date('r') . ": Executing rule " . $rule['id'] . " table " . $triggering_table . "\n";
		/* start a new messages array */
		$messages = array();
	//	$messages[] = "D: " . implode("," , $fields_to_fetch);
	//	$messages[] = "D: " . string_r($record);
	//	$messages[] = "D: " . GetBacktrace();

		/* Some initial information */
		$messages[] = "I: Start evaluation of rule '" . $rule['id'] . "' named '" . $rule['name'] . "'.";
		$messages[] = "I: Table '" . $triggering_table . "', record '" . $triggering_record . "', field '" . $triggering_field . "', condition '" . $triggering_condition . "'";

		/* Fetch conditions/conditions. Actually they are all conditions of some form, but we call them conditions. */
		$sql = "SELECT * FROM `CRMruleconditions` WHERE `CRMruleconditions`.`rule`='" . mres($rule['id']) . "' AND `CRMruleconditions`.`enabled`='Yes'";

		$conditions = db_GetArray($sql);

		if (count($conditions) > 0) {

			$operator = "AND";

			if ($rule['operator'] != "OR") {
				$operator = "AND";
			} else {
				$operator = "OR";
			}
			// Pass the list of conditions/conditions to Processruleconditions to take action if needed
			list($pass, $messages) = Processruleconditions($conditions, $record, $triggering_condition, $triggering_table, $triggering_field, $messages, $operator, $triggering_record);

			$pass_message = "";
			if (!$pass) {
				//if (is_administrator()) print( "<pre>fail: " . string_r($messages) . "</pre>\n");
				$pass_message = "FAILED CONDITIONS";
				//$messages[] = "I: Fields fetched: " . string_r($fields_to_fetch);
			} else {
				list ($pass_actions, $messages) = ProcessruleActions($triggering_table, $triggering_record, $triggering_field, $triggering_condition, $new_value, $record, $rule['id'], $messages);
				if ($pass_actions) {
					$pass_message = "PASSED CONDITIONS";
					//if (is_administrator()) print( "<pre>pass: " . string_r($messages) . "</pre>\n");
				} else {
					/* the conditions passed, but the action conditions dit not */
					$pass = false;
					$pass_message = "FAILED CONDITIONS";
				}

			}

			if (($pass && $rule['log_passed'] == "Yes") || (!$pass && $rule['log_failed'] == "Yes") || $log_because_of_development) {
				journal($rule['id'], "" . $pass_message . "\n" . implode("\n", $messages) . "", "rules");
			}
			// debug : mail("a@b.nl", "condition " . $triggering_condition, string_r($messages) . GetBackTrace());

			// The $messages array could contain user messages beginning with "U: "
			// We put them in the JavascriptToPrintAtEndOfPage_OnlyJS array first
			$messages_to_show = "";
			foreach ($messages AS $message) {
				if (substr($message, 0, 3) == "U: ") {
					$message = substr($message, 3, strlen($message) -3);
					$message = unescape_interleave_characters(ParseTemplateRecordData($message, $triggering_table, $triggering_record));
					$messages_to_show .= jsencode(html2plaintext($message)) . '\r\n';
				}
			}

			if (trim($messages_to_show) != "") {

				$messages_to_show = str_replace("\n", "\\n", $messages_to_show);
				$messages_to_show = str_replace("\r", "\\r", $messages_to_show);

				$GLOBALS['JavascriptToPrintAtEndOfPage_OnlyJS'][] = 'alert("' . trim($messages_to_show) . '");';

			}

			// It could also contain javascript, to be parsed AFTER messages
			foreach ($messages AS $message) {
				if (substr($message, 0, 3) == "JS:") {
					$message = trim(substr($message, 3, strlen($message) -3));
					$message = unescape_interleave_characters(ParseTemplateRecordData($message, $triggering_table, $triggering_record));
					$GLOBALS['JavascriptToPrintAtEndOfPage_OnlyJS'][] = $message;

				}
			}
		}
	}

	// We're done
	// Put PDS back

	$GLOBALS['UC']['PersonalDataStructure'] = $PDSPARK;
	unset($PDSPARK);


}

function Processruleconditions($conditions, $record, $triggering_condition, $table, $field, $messages, $operator="AND", $triggering_record="") {

/* conditions

	changed
	added
	deleted

	rule conditions

	[record] must be changed
	[record] must be added
	[record] must be deleted
	[field] must be changed
	[field] must be [matchvalue]
	[field] must not be [matchvalue]
	[field] must contain [matchvalue]
	[field] must not contain [matchvalue]
	[field] must be higher than [matchvalue]
	[field] must be lower than [matchvalue]
	[field] must be before [matchvalue]
	[field] must be after [matchvalue]
	[field] must be pressed [button]
	run once a day
	run once every working day
	run every minute
	run every hour
	run on sunday .. saturday
	run on first day of the month
	run on first day of the week
	system generates warning
	system generates error
	admin action request by user
	a user logs in
	group of current user must be [value] (csv)
	group of current user must not be [value] (csv)
	user id must be [value] (csv)
	user id must not be [value] (csv)

	messages array convention:
	start with "I: " means INFO
	start with "P: " means PASS
	start with "F: " means FAIL
	start with "W: " means WARNING (non-fatal)
	start with "U: " means USER : display message to the user
	start with "JS: " means JAVASCRIPT: to print at end of resulting page


*/

	// Create fully qualified field name for this table, like "t11-f123"
	$fqfield = $table . "-" . $field;

	// We're optimistic
	$pass = true;
	$some_pass = false;

	// Debug
	//print "Check condition '$triggering_condition' field '$field' operator " . $operator . "\n\n";
	//DA($conditions);

	// Cycle through all conditions/conditions to see if one happens to fail
	$messages[] = "I: Operator for conditions is '" . $operator . "'";
	foreach ($conditions AS $condition) {

		// First parse the match value, if any
		if ($condition['matchvalue'] != "") {
			//$messages[] = "I: Start with '" . $condition['matchvalue'] . "' table " . $table . " record " . $triggering_record . "";
			$condition['matchvalue'] = ParseTemplateGeneric($condition['matchvalue']);

			$condition['matchvalue'] = ParseTemplateRecordData($condition['matchvalue'], $table, $triggering_record);

			/*
			$tags = get_tags($condition['matchvalue']);
			foreach ($tags AS $tag) {
				// We only allow RAW and fully-qualified tags, like %t11-f532%
				$condition['matchvalue'] = str_replace("%" . $tag . "%", $record[$tag], $condition['matchvalue']);
			}
			*/
			if (strpos($condition['matchvalue'], "<?php") !== false) {
				$condition['matchvalue'] = EvaluateTemplatePHP($condition['matchvalue'], false, str_replace("t", "", $table), $record, false);
			}

			$condition['matchvalue'] = unescape_interleave_characters($condition['matchvalue']);
		//	$messages[] = "I: End with '" . $condition['matchvalue'] . "'";

		}
		if ($condition['condition'] == "group of current user must be [value] (csv)" || $condition['condition'] == "group of current user must not be [value] (csv)") {

			$allowed_groups = array();

			/* condition says that this rule should only fire for certain group */
			if ($condition['condition'] == "group of current user must be [value] (csv)") {
				$allowed_groups = explode(",", $condition['matchvalue']);
			} else {
				$allowed_groups = db_GetFlatArray("SELECT `id` FROM `CRMuserprofiles` WHERE `CRMuserprofiles`.`id` NOT IN ('" . implode("','", explode(",", $condition['matchvalue'])) . "')");
			}

			$messages[] = "I: Only for groups " . implode(",", $allowed_groups);

			if (in_array($GLOBALS['GROUPID'], $allowed_groups)) {
				$messages[] = "P: User is in group '" . $GLOBALS['GROUPID'] . "'.";
				$some_pass = true;
			} else {
				$messages[] = "F: User is in group '" . $GLOBALS['GROUPID'] . "'.";
				$pass = false;
			}

		} elseif ($condition['condition'] == "user id must be [value] (csv)" || $condition['condition'] == "user id must not be [value] (csv)") {

			$allowed_users = array();

			/* condition says that this rule should only fire for certain users */
			if ($condition['condition'] == "user id must be [value] (csv)") {
				$allowed_users = explode(",", $condition['matchvalue']);
			} else {
				$allowed_users = db_GetFlatArray("SELECT `id` FROM `CRMloginusers` WHERE `CRMloginusers`.`active` = 'yes' AND `CRMloginusers`.`id` NOT IN ('" . implode("','", explode(",", $condition['matchvalue'])) . "')");
			}

			$messages[] = "I: Only for users " . implode(",", $allowed_users);

			if (in_array($GLOBALS['USERID'], $allowed_users)) {
				$messages[] = "P: User is '" . $GLOBALS['USERID'] . "'.";
				$some_pass = true;
			} else {
				$messages[] = "F: User is '" . $GLOBALS['USERID'] . "'.";
				$pass = false;
			}



		} elseif ($condition['condition'] == "[record] must be added") {
			// record add (condition = "added")
			if ($triggering_condition == "added") {
				// A simple add condition of a record
				$messages[] = "P: Record must be 'added', it is according to condition";
				$some_pass = true;
			} else {
				// This record is not added, or at least we weren't called because of a change
				$pass = false;
				$messages[] = "F: Record must be 'added', but condition is '" . $triggering_condition . "'";
			}

		} elseif ($condition['condition'] == "[record] must be deleted") {
			// record deleted (condition = "deleted")
			if ($triggering_condition == "deleted") {
				// A simple delete condition of a record
				$messages[] = "P: Record must be 'deleted', it is according to condition";
				$some_pass = true;
			} else {
				// This record is not added, or at least we weren't called because of a change
				$pass = false;
				$messages[] = "F: Record must be 'deleted', but condition is '" . $triggering_condition . "'";
			}
		} elseif ($condition['condition'] == "[record] must be changed") {
			if ($field == "" || $field == "none" || $field == $table) {
				if ($triggering_condition == "changed") {
					//|| $triggering_condition == "added" || $triggering_condition == "deleted"
					// A simple change condition of a record
					$messages[] = "P: Record must be 'changed', it is according to condition (field is '" . $field . "')";
					$some_pass = true;
				} else {
					// This record is not changed, or at least we weren't called because of a change
					$pass = false;
					$messages[] = "F: Record must be 'changed', but condition is '" . $triggering_condition . "'";
				}
			} else {
				$pass = false;
				$messages[] = "F: Record must be 'changed', but condition only applies to field " . $field . ".";

			}
		} elseif ($condition['condition'] == "[field] must be changed") {

			// Change condition but now for a specific field

			if ($triggering_condition == "changed") {
				if ($fqfield == $condition['field']) {
					// A simple change condition of a field
					$messages[] = "P: Field '" . $condition['field'] . "' must be 'changed', it is according to condition.";
					$some_pass = true;
				} else {
					// This field is not changed, or at least we weren't called because of a change
					$pass = false;
					$messages[] = "F: Field '" . $condition['field'] . "' must be 'changed', the condition matches, but that is not field '" . $fqfield . "'.";
				}
			} else {
				$messages[] = "F: Record must be 'changed', but condition is '" . $triggering_condition . "'.";
				$pass = false;
			}
		} elseif ($condition['condition'] == "[field] must be [matchvalue]") {
			// Straight-on match
			if ($record[$condition['field']] != $condition['matchvalue']) {
				$pass = false;
				$messages[] = "F: Field '" . $condition['field'] . "' must be '" . $condition['matchvalue'] . "' but value '" . $record[$condition['field']] . "' is not.";
			} else {
				$messages[] = "P: Field '" . $condition['field'] . "' must be '" . $condition['matchvalue'] . "' and value '" . $record[$condition['field']] . "' is.";
				$some_pass = true;
			}
		} elseif ($condition['condition'] == "[field] must not be [matchvalue]") {
			// Mismatch
			if ($record[$condition['field']] == $condition['matchvalue']) {
				$pass = false;
				$messages[] = "F: Field '" . $condition['field'] . "' must not be '" . $condition['matchvalue'] . "' but value '" . $record[$condition['field']] . "' is.";
			} else {
				$messages[] = "P: Field '" . $condition['field'] . "' must not be '" . $condition['matchvalue'] . "' and value '" . $record[$condition['field']] . "' is not.";
				$some_pass = true;
			}

		} elseif ($condition['condition'] == "[field] must contain [matchvalue]") {
			// Must contain
			if (!stristr($record[$condition['field']], $condition['matchvalue'])) {
				$pass = false;
				$messages[] = "F: Field '" . $condition['field'] . "' must contain '" . $condition['matchvalue'] . "' but value '" . $record[$condition['field']] . "' does not.";
			} else {
				$messages[] = "P: Field '" . $condition['field'] . "' must contain '" . $condition['matchvalue'] . "' and value '" . $record[$condition['field']] . "' does.";
				$some_pass = true;
			}

		} elseif ($condition['condition'] == "[field] must not contain [matchvalue]") {
			// Must NOT contain
			if (stristr($record[$condition['field']], $condition['matchvalue'])) {
				$pass = false;
				$messages[] = "F: Field '" . $condition['field'] . "' must not contain '" . $condition['matchvalue'] . "' but value '" . $record[$condition['field']] . "' does.";
			} else {
				$messages[] = "P: Field '" . $condition['field'] . "' must not contain '" . $condition['matchvalue'] . "' and value '" . $record[$condition['field']] . "' does not.";
				$some_pass = true;
			}
		} elseif ($condition['condition'] == "[field] must be higher than [matchvalue]") {
			// Must be higher than. That excludes equal!
			if ($record[$condition['field']] <= $condition['matchvalue']) {
				$pass = false;
				$messages[] = "F: Field '" . $condition['field'] . "' must be higher than '" . $condition['matchvalue'] . "' but value '" . $record[$condition['field']] . "' is lower or equal.";
			} else {
				$messages[] = "P: Field '" . $condition['field'] . "' must be higher than '" . $condition['matchvalue'] . "' and value '" . $record[$condition['field']] . "' is higher.";
				$some_pass = true;
			}

		} elseif ($condition['condition'] == "[field] must be lower than [matchvalue]") {
			// Must be lower than. That excludes equal!
			if ($record[$condition['field']] >= $condition['matchvalue']) {
				$pass = false;
				$messages[] = "F: Field '" . $condition['field'] . "' must be lower than '" . $condition['matchvalue'] . "' but value '" . $record[$condition['field']] . "' is higher or equal.";
			} else {
				$messages[] = "P: Field '" . $condition['field'] . "' must be lower than '" . $condition['matchvalue'] . "' and value '" . $record[$condition['field']] . "' is lower.";
				$some_pass = true;
			}
		} elseif ($condition['condition'] == "[field] must be before [matchvalue]") {
			// Must be before. That excludes equal!
			if (date('U', strtotime($record[$condition['field']])) >= date('U', strtotime($condition['matchvalue']))) {
				$pass = false;
				$messages[] = "F: Field '" . $condition['field'] . "' must be before '" . $condition['matchvalue'] . "' but value '" . $record[$condition['field']] . "' is later or equal.";
				//. date('U', $record[$condition['field']]) . " >= " . date('U', $condition['matchvalue']);
			} else {
				$messages[] = "P: Field '" . $condition['field'] . "' must be before '" . $condition['matchvalue'] . "' and value '" . $record[$condition['field']] . "' is before.";
				$some_pass = true;
			}
		} elseif ($condition['condition'] == "[value] days before [field]" || $condition['condition'] == "[value] days after [field]") {
			if ($triggering_condition != $condition['condition']) {
				$pass = false;
				$messages[] = "F: Condition '" . $condition['condition'] . "' is not equal to '" . $triggering_condition . "'.";
			} else {
				$operand = "+";
				if ($condition['condition'] == "[value] days before [field]") {
					$operand = "-";
				}
	/*
	x eval
	ProcessRulesForAllTablesHavingRulesWithCondition("[value] days before [field]");
	*/

				if ($record[$condition['field']] != "" && $record[$condition['field']] != "0000-00-00") {

					$target_date = date('Y-m-d', strtotime($record[$condition['field']] . " " . $operand . $condition['matchvalue'] . " days"));

	//				$messages[] = "D: Value " . $record[$condition['field']] . " operand " . $operand . " matchvalue " . $condition['matchvalue'];

					if ($target_date != date('Y-m-d')) {

						$pass = false;

						$messages[] = "F: Field '" . $condition['field'] . "' must be '" . $record[$condition['field']] . "' " . $operand . $condition['matchvalue'] . " days (" . $target_date . ") but this is not today.";

					} else {

						$messages[] = "P: Field '" . $condition['field'] . "' must be '" . $record[$condition['field']] . "' " . $operand . $condition['matchvalue'] . " days (" . $target_date . ") and this is today.";

						$some_pass = true;
					}
				} else {

					$pass = false;

					$messages[] = "F: Field '" . $condition['field'] . "' is empty.";

				}
			}
		} elseif ($condition['condition'] == "[field] must be after [matchvalue]") {
			// Must be after. That excludes equal!
			if (date('U', strtotime($record[$condition['field']])) <= date('U', strtotime($condition['matchvalue']))) {
				$pass = false;
				$messages[] = "F: Field '" . $condition['field'] . "' must be after '" . $condition['matchvalue'] . "' but value '" . $record[$condition['field']] . "' is before or equal.";
			} else {
				$messages[] = "P: Field '" . $condition['field'] . "' must be after '" . $condition['matchvalue'] . "' and value '" . $record[$condition['field']] . "' is after.";
				$some_pass = true;
			}

		} elseif ($condition['condition'] == "[field] must be pressed [button]") {
			// ButtonPress condition
			if ($triggering_condition == "pressed") {
				if ($condition['field'] != $fqfield) {
					$pass = false;
					$messages[] = "F: Button '" . $condition['field'] . "' must be pressed but '" . $fqfield . "' was pressed.";
				} else {
					$messages[] = "P: Button '" . $condition['field'] . "' must be pressed and '" . $fqfield . "' was pressed.";
					$some_pass = true;
				}
			} else {
				$pass = false;
				$messages[] = "F: Button '" . $condition['field'] . "' must be pressed but nothing was pressed according to condition '" . $triggering_condition . "'.";
			}
		} elseif ($condition['condition'] == "run once a day" || $condition['condition'] == "run once every working day" || $condition['condition'] == "run every minute" || $condition['condition'] == "run every hour" || $condition['condition'] == "run on first day of the month" || $condition['condition'] == "run on first day of the week" || $condition['condition'] == "run on first day of the year") {

			if ($triggering_condition != $condition['condition']) {
				$pass = false;
				$messages[] = "F: Condition '" . $condition['condition'] . "' is not equal to '" . $triggering_condition . "'.";
			} else {
				$messages[] = "P: Condition '" . $condition['condition'] . "' is equal to '" . $triggering_condition . "'.";
				$some_pass = true;
			}
			/* these are all administrative mail notifications */

		} elseif ($condition['condition'] == "run on monday") {

			if ($triggering_condition != $condition['condition']) {
				$pass = false;
				$messages[] = "F: Condition '" . $condition['condition'] . "' is not equal to '" . $triggering_condition . "'.";
			} else {
				$messages[] = "P: Condition '" . $condition['condition'] . "' is equal to '" . $triggering_condition . "'.";
				$some_pass = true;
			}
		} elseif ($condition['condition'] == "run on tuesday") {

			if ($triggering_condition != $condition['condition']) {
				$pass = false;
				$messages[] = "F: Condition '" . $condition['condition'] . "' is not equal to '" . $triggering_condition . "'.";
			} else {
				$messages[] = "P: Condition '" . $condition['condition'] . "' is equal to '" . $triggering_condition . "'.";
				$some_pass = true;
			}
		} elseif ($condition['condition'] == "run on wednesday") {

			if ($triggering_condition != $condition['condition']) {
				$pass = false;
				$messages[] = "F: Condition '" . $condition['condition'] . "' is not equal to '" . $triggering_condition . "'.";
			} else {
				$messages[] = "P: Condition '" . $condition['condition'] . "' is equal to '" . $triggering_condition . "'.";
				$some_pass = true;
			}
		} elseif ($condition['condition'] == "run on thursday") {

			if ($triggering_condition != $condition['condition']) {
				$pass = false;
				$messages[] = "F: Condition '" . $condition['condition'] . "' is not equal to '" . $triggering_condition . "'.";
			} else {
				$messages[] = "P: Condition '" . $condition['condition'] . "' is equal to '" . $triggering_condition . "'.";
				$some_pass = true;
			}
		} elseif ($condition['condition'] == "run on friday") {

			if ($triggering_condition != $condition['condition']) {
				$pass = false;
				$messages[] = "F: Condition '" . $condition['condition'] . "' is not equal to '" . $triggering_condition . "'.";
			} else {
				$messages[] = "P: Condition '" . $condition['condition'] . "' is equal to '" . $triggering_condition . "'.";
				$some_pass = true;
			}
		} elseif ($condition['condition'] == "run on saturday") {

			if ($triggering_condition != $condition['condition']) {
				$pass = false;
				$messages[] = "F: Condition '" . $condition['condition'] . "' is not equal to '" . $triggering_condition . "'.";
			} else {
				$messages[] = "P: Condition '" . $condition['condition'] . "' is equal to '" . $triggering_condition . "'.";
				$some_pass = true;
			}
		} elseif ($condition['condition'] == "run on sunday") {

			if ($triggering_condition != $condition['condition']) {
				$pass = false;
				$messages[] = "F: Condition '" . $condition['condition'] . "' is not equal to '" . $triggering_condition . "'.";
			} else {
				$messages[] = "P: Condition '" . $condition['condition'] . "' is equal to '" . $triggering_condition . "'.";
				$some_pass = true;
			}

		} elseif ($condition['condition'] == "system generates warning" && $triggering_condition == "system generates warning") {

			$messages[] = "P: System warning";
			$some_pass = true;

		} elseif ($condition['condition'] == "system generates error" && $triggering_condition == "system generates error") {

			$messages[] = "P: System error";
			$some_pass = true;

		} elseif ($condition['condition'] == "admin action request by user" && $triggering_condition == "admin action request by user") {

			$messages[] = "P: Administrative action request by user";
			$some_pass = true;

		} elseif ($condition['condition'] == "a user logs in" && $triggering_condition == "a user logs in") {

			$messages[] = "P: User logging in";
			$some_pass = true;

		} elseif ($condition['condition'] == "an e-mail is received" && $triggering_condition == "an e-mail is received") {

			$messages[] = "P: Received an e-mail";
			$some_pass = true;

		} elseif ($condition['condition'] == "at the exact date and time in [field]" && $triggering_condition == "at the exact date and time in [field]") {

			if ($record[$condition['field']] != date('Y-m-d H:i:00')) {
				$pass = false;
				$messages[] = "F: Field '" . $condition['field'] . "' must be '" . date('Y-m-d H:i:00') . "' but value '" . $record[$condition['field']] . "' is not.";
			} else {
				$messages[] = "P: Field '" . $condition['field'] . "' must be '" . date('Y-m-d H:i:00') . "' and value '" . $record[$condition['field']] . "' is.";
				$some_pass = true;
			}

		} else {
			// Unknown condition
			$pass = false;
			$messages[] = "F: Condition '" . $condition['condition'] . "' is invalid.";
		}
	}
	// Return our conclusion and the messages along with it

	if ($pass == false && $some_pass == true && $operator == "OR") {
		/* this means the conditions are met if ANY of the conditions PASSED, which is the case here */
		$messages[] = "P: Conditions met because at least one condition passed and operator is 'OR'";
		$pass = true;
	}

	return array($pass, $messages);
}

function ProcessruleActions($triggering_table, $triggering_record, $triggering_field, $triggering_condition, $new_value, $record, $rule, $messages=array()) {

	// Fetch actions for this rule
	$sql = "SELECT * FROM `CRMruleactions` WHERE `CRMruleactions`.`rule`='" . mres($rule) . "' AND `CRMruleactions`.`enabled`='Yes' ORDER BY  `CRMruleactions`.`order`";
	$ruleactions = db_GetArray($sql);

	// Rule actions can fail, too. Based on their own ruleactionconditions.
	$pass = true;

	/* bool to hold flag if *something* passed */
	$some_pass = false;

	/* ordered action list */
	$todo = array();

	foreach ($ruleactions AS $action) {

		$order = $action['order'];

		$ruleactionconditions = db_GetArray("SELECT * FROM `CRMruleactionconditions` WHERE `CRMruleactionconditions`.`ruleaction`='" . mres($action['id']) . "' AND `CRMruleactionconditions`.`enabled`='Yes'");

		$action['value'] = ParseTemplateRecordData($action['value'], $triggering_table, $triggering_record);

		if ($action['alternate_from_address'] != '') {
			$action['alternate_from_address'] = unescape_interleave_characters(ParseTemplateRecordData($action['alternate_from_address'], $triggering_table, $triggering_record));
		}

		if (strpos($action['value'], "<?php") !== false) {
			$action['value'] = EvaluateTemplatePHP($action['value'], false, str_replace("t", "", $triggering_table), $triggering_record, false);
		}

		//JH 2016-04-14; is this correct? does this rule always run under the user account which triggered the rule? => HF: yes
		$action['value'] = str_replace("@USERID@", $GLOBALS['USERID'], $action['value']);
		$action['value'] = str_replace("@GROUPID@", $GLOBALS['GROUPID'], $action['value']);

		$action['value'] = unescape_interleave_characters($action['value']);

		$messages[] = "I: Start working for action '" . $action['action'] . "'.";

		if (count($ruleactionconditions) > 0) {
			$messages[] = "I: This action has its own conditions, start evaluating.";
			// This rule has its own conditions. Check them first.
			$operator = "";
			if ($action['operator'] != "OR") {
				$operator = "AND";
			} else {
				$operator = "OR";
			}

			list ($pass, $messages) = Processruleconditions($ruleactionconditions, $record, $triggering_condition, $triggering_table, $triggering_field, $messages, $operator, $triggering_record);
			if ($pass) {
				$messages[] = "P: This action had its own conditions, and they all passed.";
				$some_pass = true;
			} else {
				$messages[] = "F: This action had its own conditions, and they did not all pass. Skipping this action.";
				$pass = false;
				continue;
			}
		} else {
			$some_pass = true;
		}
		if (($action['field'] != '-' && $action['field'] != '') && !array_key_exists($action['field'], $record)) {

			$messages[] = "F: Should fetch value from " . $action['field'] . " but it is not set in result set ";

			/* action recalculate [field] misuses the field-field, so in that case, no error is nescessary */
			if ($action['action'] != "recalculate [field]") {

				log_msg("ERROR: A rule has some action set on a field (" . $action['field'] . ") which is not in the same table OR not in the fetched record (fields_to_fetch). Replace with fully qualified field name, starting with " . $triggering_table . "- etc. THIS RULE ACTION WILL FAIL. Triggering condition is " . $triggering_condition . ", rule is " . $rule['id'] . " action is " . $action['id']);

			}
		}
		$messages[] = "I: Action '" . $action['action'] . "' is appropriate.";

		if ($action['action'] == "run query in [value]") {

			$todo[$order]['queries'][] = $action['value'];
		//	$messages[] = "I: Added '" . $action['value'] . "' to list of queries to run.";

		} elseif ($action['action'] == "delete") {

			if (!is_array($todo['delete'][$triggering_table])) {
				$todo[$order]['delete'][$triggering_table] = array();
			}
			$todo[$order]['delete'][$triggering_table][] = $triggering_record;

		} elseif ($action['action'] == "recalculate current record") {

			$todo[$order]['recalcrecords'][$triggering_table][] = $triggering_record;
			$messages[] = "I: Added record '" . $triggering_record . "' in table '" . $triggering_table . "' to list of records to recalculate.";

		} elseif ($action['action'] == "recalculate [field]") {

			$todo[$order]['recalc'][] = $action['field'];
			$messages[] = "I: Added '" . $action['field'] . "' to list of tables to recalculate.";

		} elseif ($action['action'] == "mail to user(s) in [field]" || $action['action'] == "mail to address(es) in [value]" || $action['action'] == "mail to user-id(s) in [value]" || $action['action'] == "mail to group-id(s) in [value]" || $action['action'] == "mail to address in [field]"  || $action['action'] == "mail to address(es) in result of query in [value]") {

			/* dirty hack. */
			if ($triggering_condition == "admin action request by user" ) {
				$action['template'] = $new_value;
			}

			// A mail action has 3 props which make it unique: template, attach, and value. We group
			// recipients per unique combination for there cannot be any other differences.

			// Array to hold unique templates (by id)s
			if (!is_array($todo[$order]['mail'][$action['template']])) {
				$todo[$order]['mail'][$action['template']] = array();
			}

			// Determine field type of "to" field
			// Can be something like t11-f3-t2-f45

			$parts = explode("-", $action['field']);
			$target_table = $parts[count($parts)-2];
			$target_field = $parts[count($parts)-1];

			if ($action['action'] == "mail to address(es) in [value]" || $action['action'] == "mail to user-id(s) in [value]" || $action['action'] == "mail to group-id(s) in [value]" || $action['action'] == "mail to address(es) in result of query in [value]") {

				$addresses = array();

				if ($action['action'] == "mail to address(es) in result of query in [value]") {

					$addresses = db_GetFlatArray($action['value']);
					$messages[] = "I: Added '" . implode(",", $addresses) . "' to recipient list based on '" . $action['value'] . "'";

				} elseif ($action['action'] == "mail to user-id(s) in [value]") {
					//$messages[] = "D: " . $action['value'];
					$ids = explode(",", ParseTemplateGeneric($action['value']));
					$addresses = db_GetFlatArray("SELECT `CRMloginusers`.`EMAIL` FROM `CRMloginusers` WHERE `CRMloginusers`.`active`='yes' AND `CRMloginusers`.`id` IN ('" . implode("','", $ids) . "')");

				} elseif ($action['action'] == "mail to group-id(s) in [value]") {

					$ids = explode(",", ParseTemplateGeneric($action['value']));

					foreach (db_GetFlatArray("SELECT `CRMloginusers`.`EMAIL` FROM `CRMloginusers` WHERE `CRMloginusers`.`active`='yes' AND `CRMloginusers`.`PROFILE` IN ('" . implode("','", $ids) . "')") AS $recp) {
						if (ValidateEmail($recp)) {

							/* address is ok */
							$messages[] = "I: Added '" . $recp . "' to recipient list based on '" . $action['field'] . "'";
							$addresses[] = $recp;

						} elseif (ValidateEmail(GetSetting("SendundeliverableEmailsToEmailAddress"))) {

							/* the global send-failed-messages-to address */
							$messages[] = "I: Added '" . GetSetting("SendundeliverableEmailsToEmailAddress") . "' to recipient list based on SendundeliverableEmailsToEmailAddress attribute";

							$addresses[] = GetSetting("SendundeliverableEmailsToEmailAddress");


						} else {

							$messages[] = "W: Skipped adding '" . $recp . "' to recipient list based on '" . $action['field'] . "', address did not validate.";
						}

					}
				} elseif ($action['action'] == "mail to address(es) in [value]") {

					$addresses = explode(",", ParseTemplateGeneric($action['value']));

				}

				$messages[] = "I: Addresses: " . implode(",", $addresses);


				foreach ($addresses AS $recp) {

					$recp = trim($recp);

					if (ValidateEmail($recp)) {

						if (!is_array($todo[$order]['mail'][$action['template']][$recp])) {
							$todo[$order]['mail'][$action['template']][$recp] = array();
						}
						if ($action['alternate_from_address'] != "") {
							$messages[] = "I: Set '" . $action['alternate_from_address'] . "' as alternate from address.";
							$todo[$order]['mail'][$action['template']][$recp]['alternate_from_address'] = $action['alternate_from_address'];
						}

						$todo[$order]['mail'][$action['template']][$recp][$action['attach']] = $action['attach_matchvalue'];
						$messages[] = "I: Added '" . $recp . "' to recipient list based on value '" . $action['value'] . "'";

					} elseif (ValidateEmail(GetSetting("SendundeliverableEmailsToEmailAddress"))) {

						/* the global send-failed-messages-to address */
						$messages[] = "I: Added '" . GetSetting("SendundeliverableEmailsToEmailAddress") . "' to recipient list based on SendundeliverableEmailsToEmailAddress attribute";

						$recpt = GetSetting("SendundeliverableEmailsToEmailAddress");
						if (!is_array($todo[$order]['mail'][$action['template']][$recp])) {
							$todo[$order]['mail'][$action['template']][$recp] = array();
						}
						if ($action['alternate_from_address'] != "") {
							$messages[] = "I: Set '" . $action['alternate_from_address'] . "' as alternate from address.";
							$todo[$order]['mail'][$action['template']][$recp]['alternate_from_address'] = $action['alternate_from_address'];
						}

						$todo[$order]['mail'][$action['template']][$recp][$action['attach']] = $action['attach_matchvalue'];

					} else {
						$messages[] = "W: Skipped adding '" . $recp . "' to recipient list based on '" . $action['value'] . "', address did not validate.";
					}
				}



			} else {
				$DSfield = &$GLOBALS['SystemDataStructure'][$target_table]["fields"][$target_field];

				if (is_array($DSfield)) {
					// Now, process field types
					if ($DSfield['relation_to'] == "users") { // Just a plain reference to the user table
						$recp = GetUserEmail($record[$action['field']]);

						if (ValidateEmail($recp)) {

							if (!is_array( $todo[$order]['mail'][$action['template']][$recp])) $todo[$order]['mail'][$action['template']][$recp] = array();


							$todo[$order]['mail'][$action['template']][$recp][$action['attach']] = $action['value'];

							if ($action['alternate_from_address'] != "") {
								$messages[] = "I: Set '" . $action['alternate_from_address'] . "' as alternate from address.";
								$todo[$order]['mail'][$action['template']][$recp]['alternate_from_address'] = $action['alternate_from_address'];
							}

							$messages[] = "I: Added '" . $recp . "' to recipient list based on '" . $action['field'] . "'";

						} elseif (ValidateEmail(GetSetting("SendundeliverableEmailsToEmailAddress"))) {

							$recp = GetSetting("SendundeliverableEmailsToEmailAddress");

							/* the global send-failed-messages-to address */
							$messages[] = "I: Added '" . GetSetting("SendundeliverableEmailsToEmailAddress") . "' to recipient list based on SendundeliverableEmailsToEmailAddress attribute";

							if (!is_array($todo[$order]['mail'][$action['template']][$recp])) {
								$todo[$order]['mail'][$action['template']][$recp] = array();
							}
							if (!is_array($todo[$order]['mail'][$action['template']][$recp][$action['attach']])) {
								$todo[$order]['mail'][$action['template']][$recp][$action['attach']] = array();
							}

							$todo[$order]['mail'][$action['template']][$recp][$action['attach']] = $action['attach_matchvalue'];


						} else {
							$messages[] = "W: Skipped adding '" . $recp . "' to recipient list based on '" . $action['field'] . "', address did not validate.";
						}

					} elseif ($DSfield['relation_to'] == "groups") { // A reference to the groups table, add all users in the group
						foreach (db_GetFlatArray("SELECT `CRMloginusers`.`EMAIL` FROM `CRMloginusers` WHERE `CRMloginusers`.`active`='yes' AND `CRMloginusers`.`PROFILE`='" . $record[$action['field']] . "'") AS $recp) {
							if (ValidateEmail($recp)) {
								if (!is_array($todo[$order]['mail'][$action['template']][$recp])) {
									$todo[$order]['mail'][$action['template']][$recp] = array();
								}
								$todo[$order]['mail'][$action['template']][$recp][$action['attach']] = $action['attach_matchvalue'];

								if ($action['alternate_from_address'] != "") {
									$messages[] = "I: Set '" . $action['alternate_from_address'] . "' as alternate from address.";
									$todo[$order]['mail'][$action['template']][$recp]['alternate_from_address'] = $action['alternate_from_address'];
								}
								$messages[] = "I: Added '" . $recp . "' to recipient list based on '" . $action['field'] . "'";

							} elseif (ValidateEmail(GetSetting("SendundeliverableEmailsToEmailAddress"))) {

								$recp = GetSetting("SendundeliverableEmailsToEmailAddress");
								if (!is_array($todo[$order]['mail'][$action['template']][$recp])) {
									$todo[$order]['mail'][$action['template']][$recp] = array();
								}

								$todo[$order]['mail'][$action['template']][$recp][$action['attach']] = $action['attach_matchvalue'];

								if ($action['alternate_from_address'] != "") {
									$messages[] = "I: Set '" . $action['alternate_from_address'] . "' as alternate from address.";
									$todo[$order]['mail'][$action['template']][$recp]['alternate_from_address'] = $action['alternate_from_address'];
								}

								/* the global send-failed-messages-to address */
								$messages[] = "I: Added '" . GetSetting("SendundeliverableEmailsToEmailAddress") . "' to recipient list based on SendundeliverableEmailsToEmailAddress attribute";



							} else {

								$messages[] = "W: Skipped adding '" . $recp . "' to recipient list based on '" . $action['field'] . "', address did not validate.";

							}
						}
					} else {
						$recp = $record[$action['field']];
						if (ValidateEmail($recp)) {
							if (!is_array($todo[$order]['mail'][$action['template']][$recp])) {
								$todo[$order]['mail'][$action['template']][$recp] = array();
							}

							$todo[$order]['mail'][$action['template']][$recp][$action['attach']] = $action['attach_matchvalue'];

							if ($action['alternate_from_address'] != "") {
								$messages[] = "I: Set '" . $action['alternate_from_address'] . "' as alternate from address.";
								$todo[$order]['mail'][$action['template']][$recp]['alternate_from_address'] = $action['alternate_from_address'];
							}
							$messages[] = "I: Added '" . $recp . "' to recipient list based on '" . $action['field'] . "'";

						} elseif (ValidateEmail(GetSetting("SendundeliverableEmailsToEmailAddress"))) {

							$recp = GetSetting("SendundeliverableEmailsToEmailAddress");
							if (!is_array($todo[$order]['mail'][$action['template']][$recp])) {
								$todo[$order]['mail'][$action['template']][$recp] = array();
							}

							$todo[$order]['mail'][$action['template']][$recp][$action['attach']] = $action['attach_matchvalue'];

							if ($action['alternate_from_address'] != "") {
								$messages[] = "I: Set '" . $action['alternate_from_address'] . "' as alternate from address.";
								$todo[$order]['mail'][$action['template']][$recp]['alternate_from_address'] = $action['alternate_from_address'];
							}

							/* the global send-failed-messages-to address */
							$messages[] = "I: Added '" . GetSetting("SendundeliverableEmailsToEmailAddress") . "' to recipient list based on SendundeliverableEmailsToEmailAddress attribute";

						} else {
							$messages[] = "W: Skipped adding '" . $recp . "' to recipient list based on '" . $action['field'] . "', address did not validate.";
						}
					}
				} else {
					$messages[] = "F: Field '" . $action['field'] . "' could not be found in data structure (" . $target_table . "-" . $target_field . ").";
				}
			}
		} elseif ($action['action'] == "sms to number in [field]") {
			$recp = $record[$action['field']];
			if (!is_array($todo[$order]['sms'][$action['template']])) {
				$todo[$order]['sms'][$action['template']] = array();
			}
			$todo[$order]['sms'][$action['template']][] = $recp;
			$messages[] = "I: Added number '" . $recp . "' to list of numbers to SMS.";

		} elseif ($action['action'] == "sms to number(s) in [value]") {
			$numbers = explode(",", $action['value']);
			foreach ($numbers AS $recp) {
				if (!is_array($todo[$order]['sms'][$action['template']])) {
					$todo[$order]['sms'][$action['template']] = array();
				}
				$todo[$order]['sms'][$action['template']][] = $recp;
				$messages[] = "I: Added number '" . $recp . "' to list of numbers to SMS.";
			}

		} elseif ($action['action'] == "run module [module]") {
			$todo[$order]['module'][] = $action['module_to_run'];
			$messages[] = "I: Added module '" . $action['module_to_run'] . "' to list of modules to run.";

		} elseif ($action['action'] == "display template [template]") {
			$todo[$order]['template'][]= $action['template'];
			$messages[] = "I: Added template '" . $action['template'] . "' to list of templates to show (max 1).";

		} elseif ($action['action'] == "display report [template]") {
			$todo[$order]['displayreport'][] = $action['template'];
			$messages[] = "I: Added report '" . $action['template'] . "' to list of reports to show (max 1).";

		} elseif ($action['action'] == "attach report [template]") {
			$todo[$order]['attachreport'][] = $action['template'];
			$messages[] = "I: Added report '" . $action['template'] . "' to list of reports to attach to record.";

		} elseif ($action['action'] == "update value of [field] to [value]") {

			$todo[$order]['update'][$action['field']]['update'] = $action['value'];

		} elseif ($action['action'] == "update [field] to result of query in [value]") {

			$messages[] = "I: Updating to result of query '" . $action['value'] . "'";
			$todo[$order]['update'][$action['field']]['update'] = db_GetValue($action['value']);

		} elseif ($action['action'] == "add string [value] to [field]") {

			$todo[$order]['update'][$action['field']]['addstring'] = $action['value'];

		} elseif ($action['action'] == "increase [field] by [value]") {

			$todo[$order]['update'][$action['field']]['increase'] = $action['value'];

		} elseif ($action['action'] == "decrease [field] by [value]") {

			$todo[$order]['update'][$action['field']]['decrease'] = $action['value'];

		} elseif ($action['action'] == "multiply [field] by [value]") {

			$todo[$order]['update'][$action['field']]['multiply'] = $action['value'];
		}

		if ($action['message_to_display'] != "") {
			//BR" . $rule . "A" . $action['id'] . ":  eruit gehaald HF20150911
			$messages[] = "U: " . $action['message_to_display'];
		}
	}

	/*
	 * Now process all actions. The order in which stuff is done matters to a certain extend. The main order is determined by the "order" field;
	 * only actions having the same (or no) order will be processed in the perticular order below.
	 */


	foreach ($todo AS $action) {

		//$messages[] = "I: Processing: " . string_r($action);

		// COULD ALTER VALUES
		if (count($action['queries']) > 0) { // Recalculations to be performed. Always go first.
			$messages[] = "I: Start processing query actions";
			$messages = ProcessRuleQueryActions($action['queries'], $triggering_record, $triggering_table, $messages);
		}

		if (count($action['recalc']) > 0) { // Recalculations to be performed. Always go first.
			$messages[] = "I: Start processing recalc table actions";
			$messages = ProcessRuleRecalculateTableActions($action['recalc'], $triggering_record, $triggering_table, $messages);
		}

		if (count($action['recalcrecords']) > 0) {
			$messages[] = "I: Start processing recalc record actions";
			$messages = ProcessRuleRecalculateRecordActions($action['recalcrecords'], $messages);
		}

		// SHOULD ALTER VALUES
		// Update, decerase, increase, multiply, add-string, attach report

		if (count($action['update']) > 0) { // Something needs to be updated
			$messages[] = "I: Start processing update actions";
			$messages = ProcessruleUpdateActions($action['update'], $triggering_record, $triggering_table, $record, $messages);
		}

		// CANNOT ALTER VALUES but WILL ALTER THE RECORD AS A WHOLE
		if (count($action['attachreport']) > 0) { // One or more reports need to be attached
			$messages[] = "I: Start processing attach actions";
			$messages = ProcessruleReportAttachActions($action['attachreport'], $triggering_record, $triggering_table, $messages);
		}
		// DO NOT MOVE THIS! The order is vital, in any case for A65/Tsmail: FIRST attach reports, THAN run modules
		if (count($action['module']) > 0) { // Modules to run
			$messages[] = "I: Start processing module actions";
			$messages = ProcessruleModuleActions($action['module'], $triggering_record, $triggering_table, $messages);
		}
		//	// DO NOT MOVE THIS! The order is vital, in any case for A65/Tsmail

		// CANNOT ALTER VALUES
		if (count($action['mail']) > 0) { // Something needs to be mailed
			$messages[] = "I: Start processing mail actions";
			$messages = ProcessruleMailActions($action['mail'], $triggering_record, $triggering_table, $messages);
		}
		if (count($action['sms']) > 0) { // Something needs to be sms'd
			$messages[] = "I: Start processing sms actions";
			$messages = ProcessruleSMSActions($action['sms'], $triggering_record, $triggering_table, $messages);
		}
		if (count($action['displayreport']) > 0) { // A report must me popped up
			$messages = ProcessruleReportPopupActions($action['displayreport'], $triggering_record, $triggering_table, $messages);
		}

		if (count($action['template']) > 0) { // A template must be shown
			$messages[] = "I: Start processing show-template actions";
			$messages = ProcessruleTemplateActions($action['template'], $triggering_record, $triggering_table, $messages);
		}

		/* and last but not least; delete. must always come last. */
		if (count($action['delete']) > 0) {
			$messages[] = "I: Start processing delete actions";
			$messages = ProcessruleDeleteActions($action['delete'], $messages);
		}
	}
	return array($some_pass, $messages);
}
function ProcessruleDeleteActions($delete, $messages) {
	/* this function deletes all records which we concluded must be deleted */
	foreach ($delete AS $table => $records) {
		foreach ($records AS $record) {
			$return_list_only = true;
			$skip_rules = true;
			$stop = false;
			$list = delete_record_including_childs($table, $records, $return_list_only, $skip_rules);
			foreach ($list AS $tablename => $result) {
				unset($result);
				if ($tablename['nok'] > 0) {
					$stop = true;
				}
			}
			if ($stop) {
				$messages[] = "F: Record " . $record . " in table " . $table . " could not be deleted due to it having childs";
			} else {
				$return_list_only = false;
				delete_record_including_childs($table, $records, $return_list_only, $skip_rules);
				$messages[] = "P: Record " . $record . " in table " . $table . " deleted";
				$messages[] = "I: Effect: " . string_r($list);
			}

		}
	}
	return $messages;
}
function ProcessruleSMSActions($sms, $triggering_record, $triggering_table, $messages) {
	$from_name = "";
	if (GetSetting("UNIFIED_FROMADDRESS") != "") {
		$ar = ParseFormattedEmailAddress(GetSetting("UNIFIED_FROMADDRESS"));
		$from_name = $ar["fromname"];
		$messages[] = "P: 'From' override in effect for SMS (UNIFIED_FROMADDRESS): from '" . $from_name . "'";
	} else {
		$from_name = GetUserName($GLOBALS['USERID']);
		$messages[] = "P: SMS From: '" . $from_name . "'.";
	}
	// Atomos specific SMS queue table. If it doesn't exist, don't show SMS option
	$sql = "SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = 'sms'";
	if (db_GetValue($sql) == "sms") {
		$sql = "SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = 'sms' AND TABLE_NAME = 'smsqueue' LIMIT 1";
		if (db_GetValue($sql) == "smsqueue") {
			// OK we're quite positive it's OK
			foreach ($sms AS $template => $recipients) {

				$message = html2plaintext(get_parsed_template_with_given_template($triggering_table, $triggering_record, array(), array(), false, "", GetTemplateSubject($template), 'TEMPLATE_PLAIN'));

				foreach ($recipients AS $recp) {
					$result = send_sms($from_name, $recp, $message);
					if ($result) {
						$messages[] = "P: Sending SMS from '" . $from_name . "' to '" . $recp . "' message template subject '" . $template . "'.";
					} else {
						$messages[] = "F: Failed sending SMS from '" . $from_name . "' to '" . $recp . "' message template subject '" . $template . "'.";
					}
				}
			}
		}
	}
	return $messages;
}
function ProcessruleUpdateActions($update, $triggering_record, $triggering_table, $record, $messages) {
//	trace("ProcessruleUpdateActions($update, $triggering_record, $triggering_table, $record, $messages)");
//	DA($update, false);

	if (!ipi($triggering_record)) {
		/* we can't update anything without a valid record */
		$messages[] = "I: Skipping update actions, no recordid";
		return $messages;
	} else {

		/* use system data structure, we need ALL fields, not only visable */
		CheckAndCreateSystemDataStructure();

		/* it's now in the database */
		$sds = unserialize(db_GetValue("SELECT datastructure FROM CRMdatastructures WHERE userid=0 AND datastructure != ''"));

		foreach ($update AS $field_to_update => $operations) {

			$value = $record[$field_to_update];
			$old_value = $record[$field_to_update];

			foreach ($operations AS $operation => $by) {

				$by = ParseTemplateGeneric($by);
				$tags = get_tags($by);
				foreach ($tags AS $tag) {
					$by = str_replace("%" . $tag . "%", $record[$tag], $by);
				}

				if ($operation == "multiply") {
					$value = $value * $by;
					$messages[] = "P: Multiplied field '" . $field_to_update . "' by '" . $by . "' result '" . $value . "'.";
				} elseif ($operation == "increase") {
					$value = $value + $by;
					$messages[] = "P: Increased field '" . $field_to_update . "' by '" . $by . "' result '" . $value . "'.";
				} elseif ($operation == "decrease") {
					$value = $value - $by;
					$messages[] = "P: Decreased field '" . $field_to_update . "' by '" . $by . "' result '" . $value . "'.";
				} elseif ($operation == "addstring") {
					$value = $value . $by;
					$messages[] = "P: Added string to field '" . $field_to_update . "' by '" . $by . "' result '" . $value . "'.";
				} elseif ($operation == "update") {
					$value = $by;
					$messages[] = "P: Updated field '" . $field_to_update . "' to '" . $by . "'.";
				}
			}

			// Now figure out which record in which table to update
			$split = explode("-", $field_to_update);
			$target_field = $split[count($split)-1];
			$target_table = $split[count($split)-2];
			$primary = $sds[$target_table]["primary"];

			$path_to_primary = $split;
			$path_to_primary[count($path_to_primary)-1] = $primary;
			$primary_field = implode("-", $path_to_primary);

			$messages[] = "I: Target table is '" . $target_table . "', target field is '" . $target_field . "', primary field is '" . $primary_field . "' value of primary field is '" . $record[$primary_field] . "'.";

			$target_fqn = $target_table . "-" . $target_field;
			$result = "";
			$message = "";
			$returnvalue = "";

			if ($value != "") {
				list($result, $message, $returnvalue) = ValidateFieldInput(str_replace("f", "", $target_field), $value, false,true, $record[$primary_field]);
			}

			/* EXCEPTION HERE: fields updates with rules are permitted to get an empty value, even if the field is required. we do this so rules can un-set e field, in order to alert the user the next time because it's still required */

			if ($result == "nok" && $value != "") {

				/* validation failed. add message */
				$messages[] = "F: The value for '" . $field_to_update . "' did not survive the input validation test.";

				/* this is important. an automated rule returned something which didn't validate. we need to inform the administrator */
				log_msg("ERROR: A rule told us to update '" . $field_to_update . "' but the given value didn't pass the validation test. This needs to be fixed!");

				/* add any return message */
				if ($message != "") {
					$messages[] = "F: Message: " . $message;
				}

			} else {
				// Corrected value now in in $returnvalue
				if ($returnvalue == $old_value) {

					$messages[] = "W: After operation '" . $value . "' is still the same, not updating.";

				} else {
					if ($value != $returnvalue) {
						$messages[] = "I: After validation '" . $value . "' became '" . $returnvalue . "'.";
						$value = $returnvalue;
					} else {
						$messages[] = "I: After validation '" . $value . "' stayed '" . $returnvalue . "'.";
					}

					/* update_field_value wants exactly correct values. format formatted dates here */
					if ($GLOBALS['UC']['PersonalDataStructure'][$target_table]["fields"][$target_field]['fieldtype_in_extrafields_table'] == "date") {

						$value = FormattedDateToNLDate($value);
						$messages[] = "I: Formatted to NL date for database";

					} elseif ($GLOBALS['UC']['PersonalDataStructure'][$target_table]["fields"][$target_field]['fieldtype_in_extrafields_table'] == "date (iso)") {

						$value = NLDate2INTLDate(FormattedDateToNLDate($value));
						$messages[] = "I: Formatted to ISO date for database";

					} elseif ($GLOBALS['UC']['PersonalDataStructure'][$target_table]["fields"][$target_field]['ishtml']) {

                        /* EXCEPTION HERE: if the target field is a rich text field, parse it also for tags other than %-tags */
                        $value = unescape_interleave_characters(ParseTemplateRecordData($value, $triggering_table, $triggering_record));

                    }

					/* bypass security on purpose */
					$bypass_security = true;

					/* rules don't kick other rules */
					$dont_kick_rules = true;

					/* when journalling add (rule) to the comment */
					$journal_addition = "(rule)";

					/* update the value. update_field_value takes care of journaling */
					$mess = update_field_value($target_fqn, $record[$primary_field], $value, $bypass_security, $dont_kick_rules, $journal_addition);
					if ($mess != "") $messages[] = $mess;

					$messages[] = "P: Field updated " . $mess;

					//$messages[] = '<pre>' . string_r(debug_backtrace()) . '</pre>';
				}
			}
		}
		return $messages;
	}
}
function ProcessruleReportAttachActions($attachreport, $triggering_record, $triggering_table, $messages) {

	// Determine CRMbinfiles table type
	$binfiles_type = ds_tablename_to_binfiles_table($triggering_table);

	foreach ($attachreport AS $report) {
		if (ipi($report)) {

			/* generate the report */
			$pdf = GetPDFReport($report, $triggering_table, $triggering_record);

			if (strlen($pdf["pdf"]) == 0) {

				$messages[] = "F: Failed to create a pdf report";

			} else {

				$subject = "";
				$no_triggers = true;
				$is_path = false;
				$folder = false;
				$skip_security = true;
				$filename = &$pdf["filename"];
				$pdfcontent = &$pdf["pdf"];

				$file_id = AttachFile($triggering_record,$filename,$pdfcontent,$binfiles_type,"application/pdf",$subject, $no_triggers, $is_path, $folder, $skip_security);

				$messages[] = "P: Attached file '" . $filename . "' to '" . $triggering_table . "r" . $triggering_record . "' with file id '" . $file_id . "'.";
			}
		}
	}

	return $messages;
}
function ProcessruleReportPopupActions($displayreport, $triggering_record, $triggering_table, $messages) {

	$template_to_show = $displayreport[count($displayreport)-1]; // The last one wins...

    $url = "record.php?record=" . $triggering_record . "&table=" . $triggering_table . "&template_to_use=" . $template_to_show;

	$messages[] = "JS:

		window.open ('" . jsencode($url) . "');

	";

	return $messages;
}
function ProcessruleTemplateActions($template, $triggering_record, $triggering_table, $messages) {

	$template_to_show = $template[count($template)-1]; // The last one wins...

	$pageid = db_GetValue("SELECT page_id FROM CRMtemplates WHERE templateid='" . mres($template_to_show) . "'");
	$url = "";
	if ($pageid != "") {

		if (GetSetting("UseURLBeautification") != "Yes")
		{
			$url .= "page.php?";
		}
		$url .= jsencode($pageid);
		if (GetSetting("UseURLBeautification") == "Yes")
		{
			$url .= "?";
		}
		else
		{
			$url .= "&";
		}
		$url .= "table=" . $triggering_table . "&record=" . $triggering_record;
		if ($_REQUEST['nonavbar'] != "")
		{
			$url .= "&nonavbar=" . jsencode($_REQUEST['nonavbar']);
		}
		if ($_REQUEST['fromlist'] != "")
		{
			$url .= "&fromlist=" . jsencode($_REQUEST['fromlist']);
		}
	} else {
		log_msg("ERROR: A rule wants to show a template (" . $template_to_show . ") but this template doesn't have a page id! Redirecting to index.php.");
		$url = "index.php";
	}

	$messages[] = "JS:

		$(document).ready( function() {
			window.location.href = getbase() + '" . $url . "';
		});
	";
	return $messages;
}

function ProcessruleModuleActions($module, $triggering_record, $triggering_table, $messages) {
	//trace("ProcessruleModuleActions($module, $triggering_record, $triggering_table, $messages)");
	foreach ($module AS $mid) {

		$eid = false;
		$customer = false;
		$flextableid = false;
		$flextablerecord = false;

		$eid = false;
		$customer = false;
		$flextableid = str_replace("t", "", $triggering_table);
		$flextablerecord = $triggering_record;

		$return = false;
		$dqs = "";
		$st = microtime_float();
		$messages[] = "I: Start run module " . $mid . ".";

		RunModule($mid, $eid, $return, $flextablerecord, $flextableid, $customer, $dqs, $triggering_table);

		$et = microtime_float();
		$messages[] = "P: End run module " . $mid . ". Duration: " . ($et-$st) . " ms.";

	}
	return $messages;
}
function ProcessruleMailActions($mail, $triggering_record, $triggering_table, $messages) {

	// Determine CRMbinfiles table type
	$binfiles_type = ds_tablename_to_binfiles_table($triggering_table);

	$from_name = "";
	$from_mail = "";

	if (GetSetting("UNIFIED_FROMADDRESS") != "") {

		/* parse the unified from address */
		$ar = ParseFormattedEmailAddress(GetSetting("UNIFIED_FROMADDRESS"));
		$from_mail = $ar["fromemail"];
		$from_name = $ar["fromname"];
		$messages[] = "P: 'From' override in effect for mail (UNIFIED_FROMADDRESS): from '" . $from_name . "' [" . $from_mail . "] (" . GetSetting("UNIFIED_FROMADDRESS") . ")";

	} else {

		/* use the current user's details */
		$from_mail = GetUserEmail($GLOBALS['USERID']);
		$from_name = GetUserName($GLOBALS['USERID']);
		$messages[] = "P: Mail From: '" . $from_name . "' [" . $from_mail . "]";
	}

	// Process mail actions
	foreach ($mail AS $template => $mails_using_this_template) {
		$messages[] = "I: Number of mails for template " . $template . ": " . count($mails_using_this_template);
		foreach ($mails_using_this_template AS $recipient => $attachments) {
			$report_to_attach = "";
			$files_to_attach = array();

			foreach ($attachments AS $type => $matchvalue) {

				if (strstr($type, "+all")) {

                    /* +all means ALSO attach all record attachements  */
					$report_to_attach = str_replace("+all", "", $type);
					$type = "all files";

				} elseif (strstr($type, "+1") || strstr($type, "+2")) {

					$report_to_attach = $type;

				} elseif (ipi($type)) {

					$report_to_attach = $type;
				}

				//	$messages[] = "Process attachment type " . $type;

				if ($type == "all files") {

					$sql = "SELECT `CRMbinfiles`.`fileid` FROM `CRMbinfiles` WHERE `CRMbinfiles`.`koppelid`='" . $triggering_record . "' AND `CRMbinfiles`.`type`='" . $binfiles_type . "'  AND `CRMbinfiles`.`version_belonging_to`='0'";
					//print $sql . "\n";
					$files_to_attach = array_merge($files_to_attach, db_GetFlatArray($sql));

				} elseif ($type == "files matching") {

					$matchvalue = ParseTemplateGeneric($matchvalue);

					$sql = "SELECT `CRMbinfiles`.`fileid` FROM `CRMbinfiles` WHERE `CRMbinfiles`.`koppelid`='" . $triggering_record . "' AND `CRMbinfiles`.`type`='" . $binfiles_type . "' AND `CRMbinfiles`.`filename` LIKE '%" . mres($matchvalue) . "%' AND `CRMbinfiles`.`version_belonging_to`='0'";

					$files_to_attach = array_merge($files_to_attach, db_GetFlatArray($sql));

					$messages[] = "D: Matching file names against '" . $matchvalue . "', found " . count($files_to_attach) . " matching files: " . implode(",", $files_to_attach) . ". Query: " . $sql;

				}
			}

			if ($attachments['alternate_from_address'] != "") {
				$ar = ParseFormattedEmailAddress($attachments['alternate_from_address']);
				$from_mail = $ar["fromemail"];
				$from_name = $ar["fromname"];
				$messages[] = "P: 'From' override in effect for mail (alternate_from_address): from '" . $from_name . "' [" . $from_mail . "] (" . $mail['alternate_from_address'] . ")";
			}
			$messages[] = "P: Sending an e-mail to '" . $recipient . "' from '" . $from_name . "<" . $from_mail	 . ">' based on template '" . $template . "' having attachments '" . implode(",", $files_to_attach) . "' and report '" . $report_to_attach . "' parsed against record " . $triggering_record . " in table " . $triggering_table . "";

			$parse = true;

			$mail_htmltemplate = "";
			$mail_subject = "";

			/* administrative action requests don't use a template, the body is passed in the template var */
			if (ipi($template)) {

				/* RealRealMail parses the content and adds attachments */
				$mail_htmltemplate = $template;
				$parse = true;

			} else {

				$mail_subject = substr($template, 0, 50);
				$mail_htmltemplate = $template;
				$parse = false;
			}


			$flextablerecord = $triggering_record;

			$eid = false;
			$customer = false;
			$flextableid = false;

			/* TODO: these vars below? */
			$attach_to_dossier = false;
			$attach_as_filename = false;

			RealMail($mail_htmltemplate,$eid,$customer,$from_mail,$from_name,$recipient,false,$mail_subject,$attach_to_dossier,$attach_as_filename,$report_to_attach, false, $triggering_record, $parse, $files_to_attach, $triggering_table);
			$messages[] = "P: Mail sent";

		}
	}
	return $messages;
}
function ProcessRuleQueryActions($queries, $triggering_record, $triggering_table, $messages) {
	/* runs a query. no parsing for now, just run it */
	foreach ($queries AS $query) {

		$st = microtime_float();
		$messages[] = "I: Start run query.";

		mcq($query);

		$et = microtime_float();
		$messages[] = "I: End run query. Duration: " . ($et-$st) . " ms.";

	}

	return $messages;
}
function ProcessRuleRecalculateRecordActions($recalcrecords, $messages) {
	foreach ($recalcrecords AS $table => $list_of_records) {
		foreach ($list_of_records AS $record) {

			$sql = "UPDATE `CRM" . $GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'] . "` SET `CRM" . $GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'] . "`.`timestamp_last_change` = `CRM" . $GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'] . "`.`timestamp_last_change` WHERE `CRM" . $GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'] . "`.`deleted`='n' AND `CRM" . $GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'] . "`.`". $GLOBALS['UC']['PersonalDataStructure'][$table]['primary'] . "` = '" . $record . "'";

			mcq($sql);

			$messages[] = "P: Recalculated record '" . $record . "' in table '" . $table . "'";

		}

	}
	return $messages;
}
function ProcessRuleRecalculateTableActions($recalc, $triggering_record, $triggering_table, $messages) {

	/* misuse of the "field" field. we don't recalc fields, only whole tables */
	foreach ($recalc AS $fqn) {

		list ($table, $field) = explode("-", $fqn);

		$sql = "UPDATE `CRM" . $GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'] . "` SET `CRM" . $GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'] . "`.`timestamp_last_change` = `CRM" . $GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'] . "`.`timestamp_last_change` WHERE `CRM" . $GLOBALS['UC']['PersonalDataStructure'][$table]['name_in_database'] . "`.`deleted`='n'";

		$st = microtime_float();
		$messages[] = "I: Start recalulation of table '" . $table . "' (" . $st . ").";

		mcq($sql);

		$et = microtime_float();
		$messages[] = "P: End recalulation of table '" . $table . "' (" . $et . ").  Duration: " . ($et-$st) . " ms.";

	}

	return $messages;
}

function ProcessRulesForAllTablesHavingRulesWithCondition($condition) {

	// This function check all rules-per-table for a specific condition. If a rule with this condition exists, it will launch Processrules
	// This is mainly for the "once a day"-style rules.

	$sql = "SELECT COUNT(*) FROM `CRMruleconditions`, `CRMrules` WHERE `CRMruleconditions`.`rule` = `CRMrules`.`id` AND `CRMrules`.`table` IN ('none', 'users') AND `CRMruleconditions`.`condition`='" . mres($condition) . "' AND `CRMruleconditions`.`enabled` = 'Yes' AND `CRMrules`.`enabled` = 'Yes'";

	if (db_GetValue($sql) > 0) {
		/* print information when we're in cron */
		if ($GLOBALS['NIGHTLY_CRON']) print date('r') . " " . round(memory_get_usage() / 1024 / 1024, 2) . ": Working rules for table 'none'\n";
		Processrules("none", "", "", $condition, "");
	}

	/* the system data structure MUST be used because rules bypass security. The SDS is not loaded by default. */
	if (!is_array($GLOBALS['SystemDataStructure'])) {
		$GLOBALS['SystemDataStructure'] = GetSystemDataStructure();
	}

	foreach ($GLOBALS['SystemDataStructure'] AS $table => $prop) {

		/* we only do data tables (could be all tables in the future) */
		/* HF20160323 added users table */
		if ($prop['datatable'] || $table == "users") {

			/* debug */
			//print date('r') . ": Checking rules for table " . $table . "\n";

			/* fetch all rules */
			$sql = "SELECT * FROM `CRMrules` WHERE `CRMrules`.`enabled` = 'Yes' AND `CRMrules`.`table`='" . $table . "'";
			$rules = db_GetArray($sql);

			/* loop through the rules and check if they have conditions matching our given condition */

			foreach ($rules AS $rule) {

				$sql = "SELECT COUNT(*) FROM `CRMruleconditions` WHERE `CRMruleconditions`.`rule`='" . $rule['id'] . "' AND `CRMruleconditions`.`condition`='" . mres($condition) . "' AND `CRMruleconditions`.`enabled` = 'Yes'";

				/* look for and count conditions which CANNOT HAPPEN AT SERVER-SIDE */
				$sql2 = "SELECT COUNT(*) FROM `CRMruleconditions` WHERE `CRMruleconditions`.`rule`='" . $rule['id'] . "' AND `CRMruleconditions`.`enabled` = 'Yes' AND `CRMruleconditions`.`condition` IN ('[field] must be pressed [button]', '[record] must be changed', '[field] must be changed', '[record] must be added', '[record] must be deleted', 'group of current user must be [value] (csv)', 'group of current user must not be [value] (csv)', 'a user logs in')";

			//	print $sql2;
				$num_of_appliccable_conditions = db_GetValue($sql);
				$num_of_not_appliccable_conditions = db_GetValue($sql2);

				/* yep, bingo */
				if ($num_of_appliccable_conditions > 0 && $num_of_not_appliccable_conditions == 0) {

					/* debug */
					print date('r') . " " . round(memory_get_usage() / 1024 / 1024, 2) . ": Rules found for table " . $table . "\n";

					/*  should be ran for all records in the table */

					/* first, select the concerning records, if any */
					$records_sql = "SELECT `CRM" . $GLOBALS['SystemDataStructure'][$table]['name_in_database'] . "`.`" . $GLOBALS['SystemDataStructure'][$table]['primary'] . "` FROM `CRM" . $GLOBALS['SystemDataStructure'][$table]['name_in_database'] . "` WHERE ";

					if ($GLOBALS['SystemDataStructure'][$table]["logical_delete"] == true) {
						$records_sql .= "`CRM" . $GLOBALS['SystemDataStructure'][$table]['name_in_database'] . "`.`deleted`='n'";
					} elseif ($table == "users") {
						$records_sql .= " active='yes'";
					} else {
						$records_sql .= " 1=1";
					}
					/* now check if we can speed things up a little based on the conditions */
					if ($condition == "[value] days before [field]" || $condition == "[value] days after [field]") {

						$operand = "DATE_ADD";
						if ($condition == "[value] days before [field]") {
							$operand = "DATE_SUB";
						}

						/* we're going to pre-select these to avoid having to kick rules for each record */

						$sql = "SELECT * FROM `CRMruleconditions` WHERE `CRMruleconditions`.`rule`='" . $rule['id'] . "' AND `CRMruleconditions`.`condition`='" . mres($condition) . "'";

						$sql_add = array();
						$sql_text_add = "";

						/* we assume that if there's more than one condition for the same rule, it's OR. an AND operand would be useless */
						foreach (db_GetArray($sql) AS $rulecondition) {

							if ($rulecondition['field'] != "") {

								/* field names are stored FQN in rules */

								$field = str_replace($table . "-", "", $rulecondition['field']);
								$records_sql_conditions = "";

								/* add !empty for all flavors */
								if ($GLOBALS['SystemDataStructure'][$table]["fields"][$field]["dateformat_in_database"] == "dd-mm-yyyy") {

									$records_sql_conditions .= " " . "`CRM" . $GLOBALS['SystemDataStructure'][$table]['name_in_database'] . "`.`" .  $GLOBALS['SystemDataStructure'][$table]["fields"][$field]["name_in_database"] . "` != '' AND ";

								} elseif ($GLOBALS['SystemDataStructure'][$table]["fields"][$field]["dateformat_in_database"] == "yyyy-mm-dd") {

									$records_sql_conditions .= " " . "`CRM" . $GLOBALS['SystemDataStructure'][$table]['name_in_database'] . "`.`" .  $GLOBALS['SystemDataStructure'][$table]["fields"][$field]["name_in_database"] . "` != '0000-00-00' AND ";

								} elseif ($GLOBALS['SystemDataStructure'][$table]["fields"][$field]["dateformat_in_database"] == "yyyy-mm-dd hh:mm:ss") {

									$records_sql_conditions .= " " . "`CRM" . $GLOBALS['SystemDataStructure'][$table]['name_in_database'] . "`.`" .  $GLOBALS['SystemDataStructure'][$table]["fields"][$field]["name_in_database"] . "` != '0000-00-00 00:00:00' AND ";

								}

								/* operand is DATE_SUB|DATE_ADD */
								$records_sql_conditions .= $operand . "(";

								/* add correct sql syntax */
								if ($GLOBALS['SystemDataStructure'][$table]["fields"][$field]["dateformat_in_database"] == "yyyy-mm-dd") {

									$records_sql_conditions .= "`CRM" . $GLOBALS['SystemDataStructure'][$table]['name_in_database'] . "`.`" .  $GLOBALS['SystemDataStructure'][$table]["fields"][$field]["name_in_database"] . "`";

								 } elseif ($GLOBALS['SystemDataStructure'][$table]["fields"][$field]["dateformat_in_database"] == "dd-mm-yyyy") {

									$records_sql_conditions .= "STR_TO_DATE(`CRM" . $GLOBALS['SystemDataStructure'][$table]['name_in_database'] . "`.`" .  $GLOBALS['SystemDataStructure'][$table]["fields"][$field]["name_in_database"] . "`, '%d-%m-%Y')";

								 } elseif ($GLOBALS['SystemDataStructure'][$table]["fields"][$field]["dateformat_in_database"] == "yyyy-mm-dd hh:mm:ss") {

									$records_sql_conditions .= "DATE(`CRM" . $GLOBALS['SystemDataStructure'][$table]['name_in_database'] . "`.`" .  $GLOBALS['SystemDataStructure'][$table]["fields"][$field]["name_in_database"] . "`)";

								 }

								/* it's ALWAYS days in rules for these conditions */
								$records_sql_conditions .= ", INTERVAL '" . mres($rulecondition['matchvalue']) . "' DAY) = '" . date("Y-m-d") . "'";

								if ($GLOBALS['SystemDataStructure'][$table]["fields"][$field]["dateformat_in_database"] != "") {
									$sql_add[] = $records_sql_conditions;
								} else {
									log_msg("ERROR: Rule ran with condition " . $condition . " but the database name of field '" . $field . "' but this is not a date field in the datastructure. Fix this, rule '" . $rule['id'] . "'");
									print date('r') . " " . round(memory_get_usage() / 1024 / 1024, 2) . ": ERROR: Rule ran with condition " . $condition . " but the database name of field '" . $field . "' but this is not a date field in the datastructure. Fix this, rule '" . $rule['id'] . "'\n\n";
								}
							}


						}

						/* add these conditions to the main query to shorten the result hopefully */
						if (count($sql_add) > 0) {
							if ($sql_text_add != "") {
								$sql_text_add .= " AND ";
							}
							$sql_text_add .= " (" . implode(") OR (", $sql_add) . ")";
						} else {
							if ($GLOBALS['NIGHTLY_CRON']) print date('r') . " " . round(memory_get_usage() / 1024 / 1024, 2) . ": Rule " . $rule['id'] . " Tried to shorten result but something went wrong.\n";
							log_msg("ERROR: Rule " . $rule['id'] . " Tried to shorten result but something went wrong. Could it be this field is not in de DS, or not a date?");
						}
						if ($sql_text_add != "") {
							$records_sql .= " AND (" . $sql_text_add . ")";
						}

					} elseif ($condition == "at the exact date and time in [field]") {

						$sql = "SELECT * FROM `CRMruleconditions` WHERE `CRMruleconditions`.`rule`='" . $rule['id'] . "' AND `CRMruleconditions`.`condition`='" . mres($condition) . "'";
						$ruleconditions = db_GetArray($sql);

						foreach ($ruleconditions AS $rulecondition) {
							//DA($rulecondition, false);
							print date('r') . " " . round(memory_get_usage() / 1024 / 1024, 2) . ": [" . $rulecondition['condition'] . "] Rule " . $rule['id'] ." field " . $rulecondition['field'] . "\n";
							/* requires an *exact* match (resolution 1 minute) */

							if ($rulecondition['field'] != "") {
								list($dstablename, $dsfieldname) = explode("-", $rulecondition['field']);

								$records_sql .= " AND " . "`CRM" . $GLOBALS['SystemDataStructure'][$table]['name_in_database'] . "`.`" .  $GLOBALS['SystemDataStructure'][$table]["fields"][$dsfieldname]["name_in_database"] . "` = '" . date('Y-m-d H:i') . "'";
								/* debug */
								//print date('r') . ": [" . $condition . "] Rule " . $rule['id'] . " SQL: " . $records_sql . "\n";
							}
						}
					}


					/* check if we have conditions looking for an absolute (non-parsed) value */
					$sql3 = "SELECT * FROM `CRMruleconditions` WHERE `CRMruleconditions`.`rule`='" . $rule['id'] . "' AND `CRMruleconditions`.`enabled` = 'Yes' AND `CRMruleconditions`.`condition` IN ('[field] must be [matchvalue]', '[field] must not be [matchvalue]', '[field] must contain [matchvalue]', '[field] must not contain [matchvalue]', '[field] must be higher than [matchvalue]','[field] must be lower than [matchvalue]') AND `CRMruleconditions`.`matchvalue` NOT LIKE '%@%' AND `CRMruleconditions`.`matchvalue` NOT LIKE '%\%%' AND `CRMruleconditions`.`matchvalue` NOT LIKE '%<?%'";
					//print $sql3 . "\n\n";

					//DA(db_GetArray($sql3));
					foreach (db_GetArray($sql3) AS $simple_condition) {

						list($dstablename, $dsfieldname) = explode("-", $simple_condition['field']);
						$sqlcond = "";
						if ($simple_condition['condition'] == '[field] must be [matchvalue]') {
							$sqlcond = " = '" . mres($simple_condition['matchvalue']) . "'";
						} elseif ($simple_condition['condition'] == '[field] must not be [matchvalue]') {
							$sqlcond = " != '" . mres($simple_condition['matchvalue']) . "'";
						} elseif ($simple_condition['condition'] == '[field] must contain [matchvalue]') {
							$sqlcond = " LIKE '%" . mres($simple_condition['matchvalue']) . "%'";
						} elseif ($simple_condition['condition'] == '[field] must not contain [matchvalue]') {
							$sqlcond = " NOT LIKE '%" . mres($simple_condition['matchvalue']) . "%'";
						} elseif ($simple_condition['condition'] == '[field] must be higher than [matchvalue]') {
							$sqlcond = " >= '" . mres($simple_condition['matchvalue']) . "'";
						} elseif ($simple_condition['condition'] == '[field] must be lower than [matchvalue]') {
							$sqlcond = " <= '" . mres($simple_condition['matchvalue']) . "'";
						}

						$records_sql .= " AND " . "`CRM" . $GLOBALS['SystemDataStructure'][$table]['name_in_database'] . "`.`" .  $GLOBALS['SystemDataStructure'][$table]["fields"][$dsfieldname]["name_in_database"] . "` " . $sqlcond;

					}

					if ($GLOBALS['NIGHTLY_CRON']) print date('r') . " " . round(memory_get_usage() / 1024 / 1024, 2) . ": [" . $condition . "] Rule " . $rule['id'] . " SQL: " . $records_sql . "\n";
					/* fetch the list of records */
					$list_of_records = db_GetFlatArray($records_sql);

					/* send each record to processrules in a loop */
					foreach ($list_of_records AS $record) {
						/* kick processrules */
						$log_fails = false;
						if (GetSetting("DevelopmentOrProduction") == "Development") {
							$log_fails = true;
						}
						Processrules($table, $record, "", $condition, "", false, $log_fails);
					}
				} elseif ($num_of_appliccable_conditions > 0 && $num_of_not_appliccable_conditions > 0) {
						if ($GLOBALS['NIGHTLY_CRON']) print date('r') . " " . round(memory_get_usage() / 1024 / 1024, 2) . ": [" . $condition . "] Rule " . $rule['id'] . " has " . $num_of_not_appliccable_conditions . " non-appliccable conditions, skip (" . $sql2 . ")\n";
				}
			}
		}
	}
}

function consolidate_rules() {
/* consolidate. ja, lomp ja */

	$exclude = array();

	/* eerst alles met dezelfde conditions */
	foreach (db_GetArray("SELECT * FROM CRMrules") AS $rule) {
		$conds = db_GetArray("SELECT `field`,`condition`,`matchvalue`, `enabled` FROM CRMruleconditions WHERE rule='" . $rule['id'] . "'");

		foreach (db_GetArray("SELECT * FROM CRMrules WHERE id!='" . $rule['id'] . "'  AND `table`='" . $rule['table'] . "'") AS $subrule) {
			$subconds = db_GetArray("SELECT `field`,`condition`,`matchvalue`, `enabled`  FROM CRMruleconditions WHERE rule='" . $subrule['id'] . "'");
			if ($subconds == $conds) {
			//	DA($conds);
			//	DA($subconds);
				print "RULE " . $rule['id'] . " AND " . $subrule['id'] . " HAVE THE SAME CONDITIONS AND WILL BE JOINED<br>";
				mcq("UPDATE CRMruleactions SET rule='" . $rule['id'] . "' WHERE rule='" . $subrule['id'] . "'");
				mcq("UPDATE CRMrules SET remarks = CONCAT(remarks, '\n" . mres($subrule['remarks']) . "') WHERE id='" . $rule['id'] . "'");
				mcq("DELETE FROM CRMrules WHERE id='" . $subrule['id'] . "'");
				mcq("DELETE FROM CRMruleconditions WHERE rule='" . $subrule['id'] . "'");
			}
		}
	}

	/* daarna alles met dezelfde actions */

	/* alleen met 1 condition of een OR */
	foreach (db_GetArray("SELECT * FROM CRMrules WHERE (SELECT COUNT(*) FROM CRMruleconditions WHERE CRMruleconditions.rule = CRMrules.id) = 1 OR `operator`='OR'") AS $rule) {

		$actions = db_GetArray("SELECT `action`,`field`,`template`, `attach`, `value`, `module_to_run`, `attach_matchvalue`, `operator` FROM CRMruleactions WHERE rule='" . $rule['id'] . "' AND (SELECT COUNT(*) FROM CRMruleactionconditions WHERE CRMruleactionconditions.ruleaction = CRMruleactions.id) = 0");

		foreach (db_GetArray("SELECT * FROM CRMrules WHERE id!='" . $rule['id'] . "' AND `table`='" . $rule['table'] . "' AND (SELECT COUNT(*) FROM CRMruleconditions WHERE CRMruleconditions.rule = CRMrules.id) = 1 OR `operator`='OR'") AS $subrule) {

			$subactions = db_GetArray("SELECT `action`,`field`,`template`, `attach`, `value`, `module_to_run`, `attach_matchvalue`, `operator` FROM CRMruleactions WHERE rule='" . $subrule['id'] . "' AND (SELECT COUNT(*) FROM CRMruleactionconditions WHERE CRMruleactionconditions.ruleaction = CRMruleactions.id) = 0");

			if (count($subactions) > 0 && count($actions) > 0 && $subactions == $actions && $rule['id'] != $subrule['id'] && !in_array($subrule['id'], $exclude) && !in_array($rule['id'], $exclude)) {


				print "RULE " . $rule['id'] . " AND " . $subrule['id'] . " HAVE THE SAME ACTIONS AND ONLY 1 CONDITION. THEY WILL BE JOINED AND THE CONDITION OPERATOR WILL BE SET TO 'OR'<br>";
				$sql = "UPDATE CRMrules SET operator='OR' WHERE id='" . $rule['id'] . "'";
				mcq($sql);
				$sql = "UPDATE CRMruleconditions SET rule='" . $rule['id'] . "' WHERE rule='" . $subrule['id'] . "'";
				mcq($sql);
				$sql = "DELETE FROM CRMruleactions WHERE rule='" . $subrule['id'] . "'";
				mcq($sql);
				$sql = "DELETE FROM CRMrules WHERE id='" . $subrule['id'] . "'";
				mcq($sql);

				$exclude[] = $rule['id'];
				$exclude[] = $subrule['id'];

			}
		}
	}
}
?>
